@@ -56,27 +56,27 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
     def __str__(self) -> str:
         return self._path
 
-    def __eq__(self, other) -> Union[NotImplemented, bool]:
+    def __eq__(self, other) -> bool:
         if (other.__class__ is not self.__class__):
             return NotImplemented
         return self._path == other._path
 
-    def __lt__(self, other) -> Union[NotImplemented, bool]:
+    def __lt__(self, other) -> bool:
         if (other.__class__ is not self.__class__):
             return NotImplemented
         return self._path < other._path
 
-    def __le__(self, other) -> Union[NotImplemented, bool]:
+    def __le__(self, other) -> bool:
         if (other.__class__ is not self.__class__):
             return NotImplemented
         return self._path <= other._path
 
-    def __gt__(self, other) -> Union[NotImplemented, bool]:
+    def __gt__(self, other) -> bool:
         if (other.__class__ is not self.__class__):
             return NotImplemented
         return self._path > other._path
 
-    def __ge__(self, other) -> Union[NotImplemented, bool]:
+    def __ge__(self, other) -> bool:
         if (other.__class__ is not self.__class__):
             return NotImplemented
         return self._path >= other._path
@@ -292,8 +292,8 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         raise NotImplementedError
 
     @abc.abstractmethod
-    def mkdir(self, *, exist_ok: bool = False) -> None:
-        '''Create a new directory at this given path.
+    def mkdir(self: T, *, exist_ok: bool = False) -> T:
+        '''Create a new directory at this given path. Return self.
 
         If the directory already exists, and `exist_ok` is False,
         raise FileExistsError.'''
@@ -474,10 +474,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         `overwrite`: overwrite existing file?
             If False, and file exists, raises FileExistsError.
 
-        If the object pointed to is a directory, then it may or
-        may not be a problem depending on the file system.
-        For example, with a cloud blobstore, this may not be a problem.
-        With a local file system, this will raise IsADirectoryError.
+        If the object pointed to is a directory, raise IsADirectoryError.
         '''
         raise NotImplementedError
 
@@ -626,6 +623,7 @@ class LocalUpath(Upath):  # pylint: disable=abstract-method
 
     def mkdir(self, *, exist_ok=False):
         self.localpath.mkdir(parents=True, exist_ok=exist_ok)
+        return self
 
     def mv(self, target, *, overwrite=False):
         if isinstance(target, str):
@@ -689,7 +687,7 @@ class BlobUpath(Upath):  # pylint: disable=abstract-method
 
     def clear(self):
         n = 0
-        for p in self.recursive_iterdir():
+        for p in self._recursive_iterdir():
             p.rm()
             n += 1
         if n > 0:
@@ -705,7 +703,7 @@ class BlobUpath(Upath):  # pylint: disable=abstract-method
         if self._blob_exists():
             return True
         try:
-            next(self.recursive_iterdir())
+            next(self._recursive_iterdir())
             return True
         except StopIteration:
             return False
@@ -743,7 +741,7 @@ class BlobUpath(Upath):  # pylint: disable=abstract-method
         if self._as_dir:
             return True
         try:
-            next(self.recursive_iterdir())
+            next(self._recursive_iterdir())
             return True
         except StopIteration:
             if self._blob_exists():
@@ -765,7 +763,7 @@ class BlobUpath(Upath):  # pylint: disable=abstract-method
             p0 += '/'
         np0 = len(p0)
         subdirs = set()
-        for p in self.recursive_iterdir():
+        for p in self._recursive_iterdir():
             tail = p._path[np0:]
             if '/' in tail:
                 tail = tail[: tail.find('/')]
@@ -776,16 +774,31 @@ class BlobUpath(Upath):  # pylint: disable=abstract-method
     def mkdir(self, *, exist_ok=False):
         if self.is_dir():
             if exist_ok or self.is_empty_dir():
-                return
+                return self
             raise FileExistsError(self)
         else:
+            # Make sure that a path name can't be both a file
+            # and a directory.
+            p = self
+            while p._path != '/':
+                if p.is_file():
+                    raise FileExistsError(p)
+                p = p.parent
+
             self._as_dir = True
             # There is no need to "create a directory"
             # in a blob store. Just go ahead creating
             # blobs under the "directory".
+            return self
 
     @abc.abstractmethod
-    def recursive_iterdir(self: T) -> Iterator[T]:
+    def read_bytes(self):
+        if not self.is_file():
+            raise FileNotFoundError(self)
+        # subclass implementation should pick up here.
+
+    @abc.abstractmethod
+    def _recursive_iterdir(self: T) -> Iterator[T]:
         '''Yield blobs under the current "directory".
 
         For example, if self._path is
@@ -807,12 +820,39 @@ class BlobUpath(Upath):  # pylint: disable=abstract-method
         whose name starts with a given prefix.
         In this case, the prefix should be essentially
         `self._path` with '/' appended to the end.
+
+        This classes lists `_recursive_iterdir` as abstract,
+        while `iterdir` is implemented using `_recursive_iterdir`.
+        A concrete subclass may choose to implement `_recursive_iterdir`
+        (leaving `iterdir` to the implementation provided in this class),
+        or `iterdir` (and `_recursive_iterdir` implemented using `iterdir`),
+        or both `iterdir` and `_recrusive_iterdir` separately, depending
+        on the capabilities of the API of the storage engine.
         '''
         raise NotImplementedError
 
+    @abc.abstractmethod
+    def rm(self, *, missing_ok: bool = False) -> int:
+        '''Removes the file pointed to by `self`.
+        Return number of files removed.
+
+        If the file does not exist, and `missing_ok` is False,
+        raise FileNotFoundError.
+
+        If `self` is a directory, raise IsADirectoryError.
+        In this case, use `rmdir` instead.
+        '''
+        if not self.is_file():
+            if missing_ok:
+                return 0
+            if self.is_dir():
+                raise IsADirectoryError(self)
+            raise FileNotFoundError(self)
+        # subclass implementation continue from here.
+
     def rmdir(self):
         try:
-            next(self.recursive_iterdir())
+            next(self._recursive_iterdir())
             raise FileExistsError(self)
         except StopIteration:
             self._as_dir = False
@@ -823,6 +863,28 @@ class BlobUpath(Upath):  # pylint: disable=abstract-method
                exist_action: str = None) -> int:
         return self.copy_from(source, exist_action=exist_action)
 
+    @abc.abstractmethod
+    def write_bytes(self, data, *, overwrite=False):
+        # Make sure that a path name can't be both a file
+        # and a directory.
+        # TODO: this logic makes too many service calls.
+        if self.is_dir():
+            raise IsADirectoryError(self)
+        if self._path == '/':
+            raise IsADirectoryError(self)
+        p = self.parent
+        while p._path != '/':
+            if p.is_file():
+                raise FileExistsError(p)
+            p = p.parent
+
+        if self.is_file():
+            if overwrite:
+                self.rm()
+            else:
+                raise FileExistsError(self)
+        # subclass implementation should pick up here.
+
     async def a_download(self, *args, **kwargs):
         return await self._a_do(self.download, *args, **kwargs)
 
